/**
 * LK_TextEditorLib library class takes over typing functions
 * of a browser that supports javascript1.5, it handles
 * multiple TextNodes at the same time by storing text node references in a map.
 * This library is partially built out of Alex Benenson's tocOverlay.js file, the guy took
 * my idea for the cyrillic translator and the idea for selecting static text
 * in the page and converting it into cyrillic and it is only fair that I am taking some
 * of the functions, which he implemented (and improving on them too :).
 * My original implementation in the March of 2005 did not handle multiple on-screen 
 * text fields, it was quite dirty, since I only started learning heavy javascript but this library does.
 * Roman Mironenko copyright and GPL applies.
 */
var LK_TextEditorLib = {

	// this map will contain all TextNodes.
	textNodeMap : new Array(),
	isCapsOn : false,
	nodeId : 1,

	// QWERTY KEYBOARD SPECIAL CODES
	keymapSpecialFunction : new Array(59),
	keymapSpecialFunctionPressedWithShift : new Array(95),

	initTextEditorLib : function() {

		// the following setup is for Leet
		for(i=0;i<LK_TextEditorLib.keymapSpecialFunction.length;i++) {
			LK_TextEditorLib.keymapSpecialFunction[i]=false;
		}
		// removed 9, 32 (tab, space)
		var specialFunctionCodes="0,8,10,12,13,16,17,18,19,20,27,33,34,35,36,37,38,39,40,44,45,46,47,48,49,50,51,52,53,54,55,56,57,59,61,91,92,93,96".split(",");
		for(var i=0;i<specialFunctionCodes.length;i++) {
			LK_TextEditorLib.keymapSpecialFunction[parseInt(specialFunctionCodes[i])] = true;
		}
		specialFunctionCodes = null;
		
		for(i=0;i<LK_TextEditorLib.keymapSpecialFunctionPressedWithShift.length;i++) {
			LK_TextEditorLib.keymapSpecialFunctionPressedWithShift[i]=false;
		}
		var keymapSpecialFunctionPressedWithShiftCodes="41,42,43,58,60,62,63,64,94,95,123,124,125,126".split(",");
		for(var i=0;i<keymapSpecialFunctionPressedWithShiftCodes.length;i++) {
			LK_TextEditorLib.keymapSpecialFunctionPressedWithShift[parseInt(keymapSpecialFunctionPressedWithShiftCodes[i])] = true;
		}
		// Leet
	},

	turnOffAllEditors : function() {
		for (i in LK_TextEditorLib.textNodeMap) {
			var textNode = LK_TextEditorLib.textNodeMap[i];
			var attr = textNode.txtnode.getAttribute("LK_TextEditorLib.attribute.id");
			textNode.tearDown();
			delete LK_TextEditorLib.textNodeMap[attr];
			textNode.txtnode.setAttribute("LK_TextEditorLib.attribute.id",null);
		}
	},

	areAnyTextEditorsAttached : function() {
		if (!LK_TextEditorLib.isObjectEmpty(this.textNodeMap)) {
			for(var i=0;i<LK_TextEditorLib.textNodeMap.length;i++) {
				var txtNode = LK_TextEditorLib.textNodeMap[i];
				if (!LK_TextEditorLib.isObjectEmpty(txtNode)) {
					var attr = txtNode.txtnode.getAttribute("LK_TextEditorLib.attribute.id");
					if (!LK_TextEditorLib.isObjectEmpty(attr)) {
						return true;
					}
					
				}
			}
		}
		return false;
	},

	isObjectEmpty : function(obj) {
		var r;
		try {
			r= obj==undefined || obj==null || obj.toString()=="";
		} catch (e) {
			r=true;
		}
		return r;
	},

	isNodeEditor : function(node) {
		return node.getEditor != undefined || node.contentDocument != undefined;
	},

	EventTypeEventHandler : function(eventType, eventHandlerFunc) {
		this.eventType = eventType;
		this.eventHandlerFunc = eventHandlerFunc;
	},

	toggleEditor : function(func, skipSpecialCharacters, skipSpaceBar) {
		var eventHandlerArray = new Array(1);
		eventHandlerArray[0] = new LK_TextEditorLib.EventTypeEventHandler("keypress", func);
		return LK_TextEditorLib.toggleEditor1(LK_TextEditorLib.onKeyHandler, eventHandlerArray, skipSpecialCharacters, true, false, skipSpaceBar);
	},

	toggleEditorProcessAllCharCodes : function(func, skipSpecialCharacters, skipSpaceBar) {
		var eventHandlerArray = new Array(1);
		eventHandlerArray[0] = new LK_TextEditorLib.EventTypeEventHandler("keypress", func);
		return LK_TextEditorLib.toggleEditor1(LK_TextEditorLib.onKeyHandler, eventHandlerArray, skipSpecialCharacters, true, true, skipSpaceBar);
	},

	toggleEditor1 : function(onKeyHandlerFunc, eventHandlerArray, skipSpecialCharacters, attachCapsLock, processAllCharCodes, skipSpaceBar) {
		if (eventHandlerArray.length==0) {
			throw "Event handler functions undefined";
		}
		var eventType1 = eventHandlerArray[0].eventType;
		var func1 = eventHandlerArray[0].eventHandlerFunc;
		if (func1==undefined || func1==null || func1=="" || eventType1==undefined || eventType1==null || eventType1=="") {
			throw "Translation function undefined.";
		}
		var nodeTypeObject = LK_TextEditorLib.getNode();
		var node = nodeTypeObject.node;
		var isNodeRichText = nodeTypeObject.isNodeRichText;
		if (node==null) {return null;}
		var attr = node.getAttribute("LK_TextEditorLib.attribute.id");
		if (LK_TextEditorLib.isObjectEmpty(attr)) {
			attr=0;
		}
		var textNode = LK_TextEditorLib.textNodeMap[attr];
		if (textNode==undefined || LK_TextEditorLib.isObjectEmpty(attr)) {
			// define new text node
			attr = this.nodeId;
			node.setAttribute("LK_TextEditorLib.attribute.id",attr);
			this.nodeId++;
			var capsLockEventType = null;
			var capsLockFunc = null;
			if (attachCapsLock) {
				capsLockEventType = "keyup";
				capsLockFunc = LK_TextEditorLib.handleCapsLockPressed;
			}
			textNode = new LK_TextNode.TextNode(node, isNodeRichText, onKeyHandlerFunc, capsLockEventType,
				capsLockFunc, LK_TextEditorLib.isNodeEditor(node), skipSpecialCharacters, eventHandlerArray, processAllCharCodes, skipSpaceBar);
			textNode.setUp();
			LK_TextEditorLib.textNodeMap[attr]=textNode;
			return textNode;
		} else {
			// destroy existing text node
			textNode.tearDown();
			delete LK_TextEditorLib.textNodeMap[attr];
			node.setAttribute("LK_TextEditorLib.attribute.id",null);
		}
		return null;
	},

	getTextNode : function() {
		var node = LK_TextEditorLib.getNode().node;
		var attr = node.getAttribute("LK_TextEditorLib.attribute.id");
		if (LK_TextEditorLib.isObjectEmpty(attr)) {
			attr=0;
		}
		return LK_TextEditorLib.textNodeMap[attr];
	},

	NodeTypeObject : function(node, isNodeRichText) {
		this.node = node;
		this.isNodeRichText = isNodeRichText;
	},

	// try to initialize this text node with a focused node from the document.
	getNode : function() {
		var node = document.commandDispatcher.focusedElement;
		if (node==null) {
			if (document.commandDispatcher.focusedWindow) {
				if (document.commandDispatcher.focusedWindow.document) { 
					// find ThunderBird compose window (oh, yes, this is ridiculous.)
					var editors = document.getElementsByTagName("editor");
					for (var i = 0; i < editors.length; i++) {
						if (editors[i].contentWindow == document.commandDispatcher.focusedWindow) {
							return new LK_TextEditorLib.NodeTypeObject(editors[i], false);
						}
					}
				}
				node = document.commandDispatcher.focusedWindow.frameElement;
				if (node && node.contentDocument && node.contentDocument.designMode=="on") {
					return new LK_TextEditorLib.NodeTypeObject(node, true); // midas in iframe
				}
			}
		} else {
			var name = node.localName.toUpperCase();
			var type = node.type;
			if (name=="TEXTAREA" || name=="TEXTBOX" || (name=="INPUT" && (type=="text" || type=="file"))) {
				if (!node.disabled && !node.readOnly) {
					return new LK_TextEditorLib.NodeTypeObject(node, false);
				}
			}
		}
		return null;
	},


	isEnterPressed : function(key) {
		if (key==13) {
			return true;
		}
		return false;
	},

	handleCapsLockPressed : function(event) {
		if (event.which==20) {
			if (LK_TextEditorLib.isCapsOn==false) {
				LK_TextEditorLib.isCapsOn=true;
			} else {
				LK_TextEditorLib.isCapsOn=false;
			}
		}
	},

	isCapitalLetter : function(isShiftOn) {
		var isCapital=LK_TextEditorLib.isCapsOn;
		if (isShiftOn) {
			if (isCapital) {
				isCapital=false;
			} else {
				isCapital=true;
			}
		}
		return isCapital;
	},


	isEnterPressed: function (key) {
		if (key==13) {
			return true;
		}
		return false;
	},
	isSpacePressed: function (key) {
		if (key==32) {
			return true;
		}
		return false;
	},

	onKeyHandler : function(event) {
		// if this event was already processed by this handler, do not process it again
		if (event.keyCode == 255 && event.charCode > 0) {
			// I can't explain why this needs to be done, but midas will fail
			// to process event otherwise.
			if (event.target.nodeName == "HTML")
				for (var i in event) if (("" + i).toUpperCase()) event[i];
			return;
		}

		var txtnode = LK_TextEditorLib.getTextNode();
		var ch = event.which;
		if (txtnode.skipSpecialCharacters) {
			if (LK_TextEditorLib.isSpecialEditorCondition(event, ch)) {
				return;
			}
		}
		if (txtnode.skipSpaceBar) {
			if (LK_TextEditorLib.isSpacePressed(ch)) {
				return true;
			}
		}
		if (txtnode.processAllCharCodes || (event.charCode > 0 && !event.ctrlKey && !event.altKey && !event.metaKey)) {
			// is shift on, is typed letter capital?
			var isShiftOn = event.shiftKey;
			var isCapital=LK_TextEditorLib.isCapitalLetter(isShiftOn);
			// translate typed character
			var translator = txtnode.translationFunction;
			var ch = translator(ch, isCapital, isShiftOn, event);
			if (ch==-1) {
				return;
			}

			event.preventDefault();
			// in case if the return value is an array, print each element of the array in sequence.
			if (ch instanceof Array) {
				for(var i=0;i<ch.length;i++) {
					if (ch[i]=="\8") {
						LK_TextEditorLib.forwardToTarget(event,event.DOM_VK_BACK_SPACE,ch[i]);
					} else {
						LK_TextEditorLib.forwardToTarget(event,255,ch[i]);
					}
				}
			} else {
				LK_TextEditorLib.forwardToTarget(event,255,ch);
			}
		}
	},

	// forward to the target of the event the new chars.
	forwardToTarget : function (event,eventType,ch){
			var newEvent = document.createEvent("KeyboardEvent");
			newEvent.initKeyEvent(event.type, event.bubbles, event.cancelable, event.view, false, false, false, false, eventType, ch);
			event.target.dispatchEvent(newEvent);
	},


	onKeyReverseHandler : function(event) {
		var txtnode = LK_TextEditorLib.getTextNode();
		// if this event was already processed by this handler, do not process it again
		if ((event.keyCode == 255 && event.charCode > 0) ||
				(event.keyCode==46 && event.charCode==1 && txtnode.prevKeyCode==37) ||
				(event.keyCode==8 && event.charCode==8 && txtnode.prevKeyCode==39) ||
				((event.keyCode == 37 || event.keyCode == 39) && (event.charCode ==1 || event.charCode==2))) {
			if (event.keyCode==39 && event.charCode==1) {
				txtnode.prevKeyCode=39;
			} else if (event.keyCode==37 && event.charCode==2) {
				txtnode.prevKeyCode=37;
			}
			// I can't explain why this needs to be done, but midas will fail
			// to process event otherwise.
			if (event.target.nodeName == "HTML")
				for (var i in event) if (("" + i).toUpperCase()) event[i];
			return;
		}
		var ch = event.which;
		txtnode.prevKeyCode=null;

		// handle backspace
		if (ch==8) {
			event.preventDefault();
			var newEvent = document.createEvent("KeyEvents");
			newEvent.initKeyEvent(event.type, event.canBubble, event.cancelable, event.view, false, false, false, false, 39, 1);
			event.target.dispatchEvent(newEvent);
			newEvent = document.createEvent("KeyEvents");
			newEvent.initKeyEvent(event.type, event.canBubble, event.cancelable, event.view, false, false, false, false, 8, 8);
			event.target.dispatchEvent(newEvent);
			return;
		}
		// handle delete
		if (event.keyCode==46) {
			event.preventDefault();
			LK_TextEditorLib.sendTwoKeyEvents(event, 37, 2, 46, 1);
			return;
		}

		if (txtnode.skipSpecialCharacters) {
			if (LK_TextEditorLib.isSpecialEditorCondition(event, ch)) {
				return;
			}
		}
		if (txtnode.skipSpaceBar) {
			if (LK_TextEditorLib.isSpacePressed(ch)) {
				return true;
			}
		}
		if (txtnode.processAllCharCodes || (event.charCode > 0 && !event.ctrlKey && !event.altKey && !event.metaKey)) {

			// is shift on, is typed letter capital?
			var isShiftOn = event.shiftKey;
			var isCapital=LK_TextEditorLib.isCapitalLetter(isShiftOn);
			// translate typed character
			var translator = txtnode.translationFunction;
			var ch = translator(ch, isCapital, isShiftOn, event);
			if (ch==-1) {
				return;
			}
			event.preventDefault();
			// in case if the return value is an array, print each element of the array in sequence.
			if (ch instanceof Array) {
				for(var i=0;i<ch.length;i++) {
					LK_TextEditorLib.sendTwoKeyEvents(event, 255, ch[i], 37, 1);
				}
			} else {
				LK_TextEditorLib.sendTwoKeyEvents(event, 255, ch, 37, 1);
			}
		}
	},

	sendTwoKeyEvents : function(event, keyCode1, charCode1, keyCode2, charCode2) {
			var newEvent = document.createEvent("KeyEvents");
			newEvent.initKeyEvent(event.type, event.canBubble, event.cancelable, event.view, false, false, false, false, keyCode1, charCode1);
			event.target.dispatchEvent(newEvent);
			newEvent = document.createEvent("KeyEvents");
			newEvent.initKeyEvent(event.type, event.canBubble, event.cancelable, event.view, false, false, false, false, keyCode2, charCode2);
			event.target.dispatchEvent(newEvent);
	},

	isSpecialEditorCondition : function (event, ch) {
		if (LK_TextEditorLib.isEnterPressed()) {
			return true;
		}
		if (LK_TextEditorLib.keymapSpecialFunction[ch]==true) {
			return true;
		}
		if (event.ctrlKey || event.altKey || event.shiftKey && LK_TextEditorLib.keymapSpecialFunctionPressedWithShift[ch]) {
			return true;
		}
		return false;
	},

	onKeyHandlerTranslateToStringNoEventOverriding : function(event) {
		var txtnode = LK_TextEditorLib.getTextNode();
		var ch = event.which;
		if (txtnode.skipSpecialCharacters) {
			if (LK_TextEditorLib.isSpecialEditorCondition(event, ch)) {
				return;
			}
		}
		if (txtnode.skipSpaceBar) {
			if (LK_TextEditorLib.isSpacePressed(ch)) {
				return true;
			}
		}
		if (txtnode.processAllCharCodes || event.which > 0 || (event.charCode > 0 && !event.ctrlKey && !event.altKey && !event.metaKey)) {
			// is shift on, is typed letter capital?
			var isShiftOn = event.shiftKey;
			var isCapital=LK_TextEditorLib.isCapitalLetter(isShiftOn);
			// translate typed character
			var translator = txtnode.translationFunction;
			var str = translator(ch, isCapital, isShiftOn, event);
			if (str==-1) {
				return;
			}
			event.preventDefault();
			if (txtnode.isNodeRichText) {
				var focused = document.commandDispatcher.focusedElement;
				if (focused == null) {
					focused = document.commandDispatcher.focusedWindow;
				}
				var selection = focused.getSelection();
				newNode = document.createTextNode(str);
				LK_TextEditorLib.insertNodeAtSelection(selection, newNode);
			} else {
				var target = event.target;
				var fontwidth=8;
				var fontheight=16;
				var cursorPos = target.selectionStart;
				var origYpos = target.scrollTop;
				var origXpos = target.scrollLeft;
				// get text before the new character
				var textbefore = target.value.substring(0, target.selectionStart);
				// set new value to the textarea
				target.value = textbefore + str + target.value.substring(target.selectionEnd);
				// get all text before new char including new char to set cursor
				var textfromstart = textbefore+str;
				// set cursor after the new character
				target.setSelectionRange(textfromstart.length, textfromstart.length);
				// record the height of the text before and including the new char by counting lines * line height
				var newTextLinesHeight = (textfromstart.split("\n").length)*(fontheight);
				// record the x position at the cursor position after the new char
				var newTextLineWidth = (textfromstart.length-textfromstart.lastIndexOf("\n"))*(fontwidth);
				// record the total height of the text (number of lines * font height)
				var totalTextHeight = (target.rows)*(fontheight);
				// record the total width of the text input field
				var totalTextAreaWidth = target.clientWidth;
				// typing in a text area or smth else (input type text)?
				var isTextArea = false;
				if (totalTextHeight) {
					isTextArea = true;
				}
				if (isTextArea) {	
					// fix a problem that happens when a text box is expanded (blown up) by
					// a single word that is longer than the text box max width.
					// In this case cursor starts jumping in Russ Key mode.
					// find the longest sentence in the text box and compare the length of that sentence
					// to the total text area width.  The bigger number wins.
					var lines = target.value.split("\n");
					var longestword = 0;
					for (var line in lines) {
						var words = lines[line].split(" ");
						for (var word in words) {
							if (words[word].length > longestword) {
								longestword = words[word].length;
							}
						}
					}
					if (longestword > totalTextAreaWidth) {
						totalTextAreaWidth = longestword;
					}
				
					if ((newTextLinesHeight>origYpos)&&(newTextLinesHeight<(origYpos+totalTextHeight))) {
						//if the line where the new char is the same as before typing,
						//scroll to the original line
						target.scrollTop=origYpos;
					} else if (newTextLinesHeight<=origYpos) {
						//if the line where the new char is above the original line,
						//subtract height of one line.
						target.scrollTop=newTextLinesHeight-(fontheight);
					} else if (newTextLinesHeight>=(origYpos+totalTextHeight)) {
						//if the line with the new char is below the visible area
						//of the input field set line to the last row
						target.scrollTop=newTextLinesHeight-totalTextHeight;
					}
					if ((newTextLineWidth>=origXpos)&&(newTextLineWidth<(origXpos+totalTextAreaWidth))) {
						//if the new character in line is greater than the original position
						//and yet less than the total text input field width, scroll to the
						// original position (actually this does nothing.)
						target.scrollLeft=origXpos;
					} else if (newTextLineWidth<origXpos) {
						//if the new x position is less than the original position
						//scroll to the new line less one character
						target.scrollLeft=newTextLineWidth-(fontwidth);
					} else if (newTextLineWidth>=(origXpos+totalTextAreaWidth)) {
						//if the new character position is greater or equal to the width
						//of the text input field, scroll to the new character less the width
						//of the text input field (this means that if the text input field is 100
						//characters and the new char is at 101, the field will scroll to 101 and
						// only character 101 will be seen at the screen position 1.
						target.scrollLeft=newTextLineWidth-totalTextAreaWidth;
					}
				} else {
					// Typing in smth else than TEXT AREA (INPUT TYPE TEXT)
					target.scrollLeft=newTextLineWidth-totalTextAreaWidth;
				}
			}
		}
	},
/*
	onKeyReverseHandlerTranslateToStringNoEventOverriding : function(event) {
		var txtnode = LK_TextEditorLib.getTextNode();
		var ch = event.which;
		if (txtnode.skipSpecialCharacters) {
			if (LK_TextEditorLib.isSpecialEditorCondition(event, ch)) {
				return;
			}
		}
		if (txtnode.skipSpaceBar) {
			if (LK_TextEditorLib.isSpacePressed(ch)) {
				return true;
			}
		}
		if (txtnode.processAllCharCodes || event.which > 0 || (event.charCode > 0 && !event.ctrlKey && !event.altKey && !event.metaKey)) {
			// is shift on, is typed letter capital?
			var isShiftOn = event.shiftKey;
			var isCapital=LK_TextEditorLib.isCapitalLetter(isShiftOn);
			// translate typed character
			var translator = txtnode.translationFunction;
			var str = translator(ch, isCapital, isShiftOn, event);
			if (str==-1) {
				return;
			}
			event.preventDefault();
			if (txtnode.isNodeRichText) {
				var focused = document.commandDispatcher.focusedElement;
				if (focused == null) {
					focused = document.commandDispatcher.focusedWindow;
				}
				var selection = focused.getSelection();
				newNode = document.createTextNode(str);
				LK_TextEditorLib.insertNodeAtSelection(selection, newNode);
			} else {
				var target = event.target;
				var fontwidth=8;
				var fontheight=16;
				var cursorPos = target.selectionStart;
				var origYpos = target.scrollTop;
				var origXpos = target.scrollLeft;
				// get text before the new character
				var textbefore = target.value.substring(0, target.selectionStart);
				// set new value to the textarea
				target.value = textbefore + str + target.value.substring(target.selectionEnd);
				// get all text before new char including new char to set cursor
				var textfromstart = textbefore+str;
				// set cursor after the new character
				target.setSelectionRange(textfromstart.length, textfromstart.length);
				// record the height of the text before and including the new char by counting lines * line height
				var newTextLinesHeight = (textfromstart.split("\n").length)*(fontheight);
				// record the x position at the cursor position after the new char
				var newTextLineWidth = (textfromstart.length-textfromstart.lastIndexOf("\n"))*(fontwidth);
				// record the total height of the text (number of lines * font height)
				var totalTextHeight = (target.rows)*(fontheight);
				// record the total width of the text input field
				var totalTextAreaWidth = target.clientWidth;
				// typing in a text area or smth else (input type text)?
				var isTextArea = false;
				if (totalTextHeight) {
					isTextArea = true;
				}
				if (isTextArea) {	
					// fix a problem that happens when a text box is expanded (blown up) by
					// a single word that is longer than the text box max width.
					// In this case cursor starts jumping in Russ Key mode.
					// find the longest sentence in the text box and compare the length of that sentence
					// to the total text area width.  The bigger number wins.
					var lines = target.value.split("\n");
					var longestword = 0;
					for (var line in lines) {
						var words = lines[line].split(" ");
						for (var word in words) {
							if (words[word].length > longestword) {
								longestword = words[word].length;
							}
						}
					}
					if (longestword > totalTextAreaWidth) {
						totalTextAreaWidth = longestword;
					}
				
					if ((newTextLinesHeight>origYpos)&&(newTextLinesHeight<(origYpos+totalTextHeight))) {
						//if the line where the new char is the same as before typing,
						//scroll to the original line
						target.scrollTop=origYpos;
					} else if (newTextLinesHeight<=origYpos) {
						//if the line where the new char is above the original line,
						//subtract height of one line.
						target.scrollTop=newTextLinesHeight-(fontheight);
					} else if (newTextLinesHeight>=(origYpos+totalTextHeight)) {
						//if the line with the new char is below the visible area
						//of the input field set line to the last row
						target.scrollTop=newTextLinesHeight-totalTextHeight;
					}
					if ((newTextLineWidth>=origXpos)&&(newTextLineWidth<(origXpos+totalTextAreaWidth))) {
						//if the new character in line is greater than the original position
						//and yet less than the total text input field width, scroll to the
						// original position (actually this does nothing.)
						target.scrollLeft=origXpos;
					} else if (newTextLineWidth<origXpos) {
						//if the new x position is less than the original position
						//scroll to the new line less one character
						target.scrollLeft=newTextLineWidth-(fontwidth);
					} else if (newTextLineWidth>=(origXpos+totalTextAreaWidth)) {
						//if the new character position is greater or equal to the width
						//of the text input field, scroll to the new character less the width
						//of the text input field (this means that if the text input field is 100
						//characters and the new char is at 101, the field will scroll to 101 and
						// only character 101 will be seen at the screen position 1.
						target.scrollLeft=newTextLineWidth-totalTextAreaWidth;
					}
				} else {
					// Typing in smth else than TEXT AREA (INPUT TYPE TEXT)
					target.scrollLeft=newTextLineWidth-totalTextAreaWidth;
				}
			}
		}
	},
*/
	// transformer functions

	getSelection : function(node) {
		if (node==null) {
			// retrieve global selection
			var window = document.commandDispatcher.focusedWindow;
			if (window && window.getSelection) {
				return window.getSelection();
			}
		} else {
			// return selection from editor or from midas content window
			return node.getSelection ? node.getSelection() : node.contentWindow ? node.contentWindow.getSelection() : null;
		}
		return null;
	},

	splitHtmlString: function(string) {
		var re = /<[\/]?[!A-Z][^>]*>/ig;
		var result = new Array();
		var lastIndex = 0;
		var arr = null;
		while ( (arr = re.exec(string)) != null) {
			result[result.length] = string.substring(lastIndex, arr.index);
			result[result.length] = string.substring(arr.index, re.lastIndex);
			lastIndex = re.lastIndex;
		}
		result[result.length] = string.substr(lastIndex);
		return result;
	},

	convertWithHTML: function(src, skipHtml, converter) {
		if (src == "" || src == null) return src;
		if (!skipHtml) {
			return converter(src);
		} else {
			var arr = LK_TextEditorLib.splitHtmlString(src);
			for (var i = 0; i < arr.length; i++) {
				if ( (i % 2) == 0) arr[i] = converter(arr[i]);
			}
			return arr.join("");
		}
	},

	insertNodeAtSelection : function(selection, insertNode)
	{
		var range = selection.getRangeAt(0);
		selection.removeAllRanges();
		range.deleteContents();
		var container = range.startContainer;
		var pos = range.startOffset;
		if (!pos) {
			var newNode = document.createTextNode(" ");
			range.insertNode(newNode);
			selection.addRange(range);
			range = selection.getRangeAt(0);
			selection.removeAllRanges();
			range.deleteContents();
			container = range.startContainer;
			pos = range.startOffset;
		}
		range=document.createRange();
		// insertion logic
		// special case inserting text into text node
		if (container.nodeType==insertNode.TEXT_NODE && insertNode.nodeType==insertNode.TEXT_NODE) {
			container.insertData(pos, insertNode.nodeValue);
			// set cursor position at the end
			range.setEnd(container, pos+insertNode.length);
			range.setStart(container, pos+insertNode.length);
		} else {
			var afterNode;
			if (container.nodeType==insertNode.TEXT_NODE) {
				// inserting into a textnode, create 2 new nodes
				// insert the new node inbetween
				var textNode = container;
				container = textNode.parentNode;
				var text = textNode.nodeValue;
				// text before the split
				var textBefore = text.substr(0,pos);
				// text after the split
				var textAfter = text.substr(pos);
				var beforeNode = document.createTextNode(textBefore);
				afterNode = document.createTextNode(textAfter);
				// insert the all new nodes before the old one
				container.insertBefore(afterNode, textNode);
				container.insertBefore(insertNode, afterNode);
				container.insertBefore(beforeNode, insertNode);
				// remove the old node
				container.removeChild(textNode);
			} else {
				// else simply insert the node
				afterNode = container.childNodes[pos];
				container.insertBefore(insertNode, afterNode);
			}
			// set cursor
			range.setEnd(afterNode, 0);
			range.setStart(afterNode, 0);
		}
		selection.addRange(range);
	},

	isTransformableNodeType : function(node) {
		return node.nodeType == node.TEXT_NODE || node.nodeType == node.PROCESSING_INSTRUCTION_NODE || node.nodeType == node.COMMENT_NODE;
	},

	isBeginningOfTransformableContainer : function(range, node, state) {
		var startContainer = range.startContainer;
		if (!state.started && 
			((LK_TextEditorLib.isTransformableNodeType(startContainer) && node == state.range.startContainer) ||
					(startContainer.childNodes && node == startContainer.childNodes[range.startOffset]))) {
			return true;
		}
		return false;
	},

	isEndOfTransformableContainer : function(range, node, state) {
		var endContainer = range.endContainer;
		if (!state.finished &&
			((LK_TextEditorLib.isTransformableNodeType(endContainer) && node == endContainer) || ((endContainer.childNodes.length > 0) &&
					node == endContainer.childNodes[state.range.endOffset - 1]))) {
			return true;
		}
		return false;
	},

	NodeRangeConversionState : function(state, node) {
		this.node = node;
		this.range = state.range;
		this.convert = state.convert;
		this.isAsynchronous = state.isAsynchronous;
		this.started = state.started;
		this.finished = state.finished;
		this.converted = state.converted;
		this.toString = function() {
			return "started : " + this.started + ", finished: " + this.finished;
		};
	},

	transformNodeText : function(node, state) {
		var start = (node == state.range.startContainer) ? state.range.startOffset : 0;
		var end   = (node == state.range.endContainer) ? state.range.endOffset : node.nodeValue.length;
		var remainder = (node == state.range.endContainer) ? node.nodeValue.length - state.range.endOffset : 0;
		var convertedValue;
		if (state.isAsynchronous) {
			var stateCopy = new LK_TextEditorLib.NodeRangeConversionState(state, node);
			// call asynchronous conversion function, pass it the text to be converted and
			// copy of the state, so that the conversion function can initiate conversion, return
			// and at the end of conversion make a call back to this function with the copied state
			// in synchronous mode.
			var conversionResult = state.convert(node.nodeValue.substring(start, end), stateCopy);
			// if conversionResult is not null, then the call was resolved synchronously (from cache,)
			// and there will be no asynchronous call back.
			if (conversionResult!=null) {
				convertedValue = node.nodeValue.substring(0, start) + conversionResult + node.nodeValue.substr(end);
			}
		} else {
			// call synchronous conversion function.
			convertedValue = node.nodeValue.substring(0, start) + state.convert(node.nodeValue.substring(start, end)) + node.nodeValue.substr(end);
		}
		state.converted = true;
		if (!LK_TextEditorLib.isObjectEmpty(convertedValue)) {
			node.nodeValue = convertedValue;
			if (node == state.range.endContainer) {
				state.range.setEnd(node, node.nodeValue.length - remainder);
			}
			if (node == state.range.startContainer) {
				state.range.setStart(node, start);
			}
		}
	},

	RangeConversionState: function(range, converter, isAsynchronous) {
		this.range = range;
		this.convert = converter;
		this.isAsynchronous = isAsynchronous;
		this.started = false;
		this.finished = false;
		this.converted = false;
		this.toString = function() {
			return "started : " + this.started + ", finished: " + this.finished;
		};
	},

	transformRangeNode : function(node, state) {
		if (state.started && state.finished) {
			return;
		}
		if (LK_TextEditorLib.isBeginningOfTransformableContainer(state.range, node, state)) {
			state.started = true;
		}
		if (LK_TextEditorLib.isTransformableNodeType(node)) {
			if (state.started && !state.finished) {
				LK_TextEditorLib.transformNodeText(node, state);
			}
		} else if (node.childNodes) {
			for (var i = 0; i < node.childNodes.length; i++) {
				LK_TextEditorLib.transformRangeNode(node.childNodes[i], state);
				if (state.started && state.finished)
				break;
			}
		}
		if (LK_TextEditorLib.isEndOfTransformableContainer(state.range, node, state)) {
			state.finished = true;
		}
	},

	transformHTMLSelection : function(selection, transformerFunc, isAsynchronous) {
		if (selection==null || LK_TextEditorLib.isObjectEmpty(transformerFunc)) {
			return;
		}
		for (var i=0;i<selection.rangeCount;i++) {
			var conversionState = new LK_TextEditorLib.RangeConversionState(selection.getRangeAt(i), transformerFunc, isAsynchronous);
			LK_TextEditorLib.transformRangeNode(selection.getRangeAt(i).commonAncestorContainer, conversionState);
			if (!conversionState.converted && selection.getRangeAt(i).toString().length>0) {
				// extra conversion handling
				var range = selection.getRangeAt(i);
				var newNode = document.createTextNode(transformerFunc(range.toString()));
				LK_TextEditorLib.insertNodeAtSelection(selection, newNode);
			}
		}
		
	},

	transformHTMLSelectionAsSingleRange : function(selection, transformerFunc, isAsynchronous) {
		if (selection==null || LK_TextEditorLib.isObjectEmpty(transformerFunc)) {
			return;
		}
		var allText = "";
		for (var i=0;i<selection.rangeCount;i++) {
			allText += selection.getRangeAt(i).toString() + "\r";
		}
		var newNode = document.createTextNode(transformerFunc(allText));
		LK_TextEditorLib.insertNodeAtSelection(selection, newNode);		
	},

	transformSelectionText : function(transformerFunc, skipHtml, isAsynchronous) {
		var nodeTypeObj = LK_TextEditorLib.getNode();
		var node = nodeTypeObj==null ? null : nodeTypeObj.node;
		if (node!=null || !LK_TextEditorLib.isObjectEmpty(LK_TextEditorLib.getSelection(null))) {
			if (node==null || LK_TextEditorLib.isNodeEditor(node)) {
				// transform text in an HTML document or HTML editor
				LK_TextEditorLib.transformHTMLSelection(LK_TextEditorLib.getSelection(node), transformerFunc, isAsynchronous);				
			} else if (node!=null) {
				// transform text in a text input field or textarea
				var oldValue = node.value;
				var selStart = node.selectionStart;
				var selEnd = node.selectionEnd;
				var value = oldValue.substring(selStart, selEnd);
				var newValue = LK_TextEditorLib.convertWithHTML(value, skipHtml, transformerFunc);
				var scrollTop = node.scrollTop;
				node.value = oldValue.substring(0, selStart) + newValue + oldValue.substring(selEnd, oldValue.length);
				node.selectionStart = selStart + value.length;
				node.selectionEnd = selStart + value.length;
			}
		}
	},

	//tabEditorDocumentMap : new Array(),

	createNewBrowserTabEditor : function(title,name,id,cols,rows,wrap,initvalue,selected) {
		var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService(Components.interfaces.nsIWindowMediator);
		var printWindow = wm.getMostRecentWindow("navigator:browser");
		//var doc = printWindow.gBrowser.contentDocument;
		if (LK_TextEditorLib.isObjectEmpty(printWindow)) {
			return;
		}
		printWindow.gBrowser.selectedTab = printWindow.gBrowser.addTab("about:blank");
		// thanks, grimholtz, you had to suffer with this first in your PasswordMaker, and thanks gorgias :)
		printWindow.gBrowser.selectedBrowser.addEventListener("load", function() {
			var doc = printWindow.gBrowser.contentDocument;
			if (doc.location=="about:blank") {
				// remove document content
				doc.removeChild(doc.getElementsByTagName("html")[0]);
				// create new content
				var dom = new DOMParser().parseFromString('<html><body><div id="lk_editor_page"></div></body></html>',"application/xml");
				doc.appendChild(dom.documentElement);
				var divNode=doc.getElementsByTagName("div").item(0);
				divNode.appendChild(doc.createTextNode(title));
				divNode.appendChild(doc.createElement("br"));
				var txta = doc.createElement("textarea");
				txta.setAttribute("id", id);
				txta.setAttribute("name",name);
				txta.setAttribute("cols",cols);
				txta.setAttribute("rows",rows);
				txta.setAttribute("wrap",wrap);
				divNode.appendChild(txta);
				txta.value=initvalue;
				txta.focus();
				if (selected) {
					txta.setSelectionRange(0,initvalue.length);
				}
				//var l = LK_TextEditorLib.tabEditorDocumentMap.length+1;
				//LK_TextEditorLib.tabEditorDocumentMap[l]=doc;
			}
		}, true);
	}

}

/**
 * TextNode class is used to hold reference to a text editor field.
 * TextNode can initialize itself from a document node, it handles
 * event listeners on the editor field.
 */

var LK_TextNode = {
	//txtnode : undefined,
	//eventToListen : undefined,
	//translationFunction : undefined,
	bgcOffsetRed : 0,
	bgcOffsetGreen : 0,
	bgcOffsetBlue : 0,
	fgcOffsetRed : 0,
	fgcOffsetGreen : 0,
	fgcOffsetBlue : 0,

	TextNode : function(node, isNodeRichText, onKeyHandlerFunc, onCapsLockEventToListen, onCapsLockHandlerFunc, isNodeEditor,
			skipSpecialCharacters, eventHandlerArray, processAllCharCodes, skipSpaceBar) {
		this.txtnode = node;
		this.isNodeRichText = isNodeRichText;
		this.eventToListen = eventHandlerArray[0].eventType;
		this.onKeyHandlerFunc = onKeyHandlerFunc;
		this.translationFunction = eventHandlerArray[0].eventHandlerFunc;
		this.onCapsLockEventToListen = onCapsLockEventToListen;
		this.onCapsLockHandlerFunc = onCapsLockHandlerFunc;
		this.isNodeEditor = isNodeEditor;
		this.skipSpecialCharacters = skipSpecialCharacters;
		this.eventHandlerArray = eventHandlerArray;
		this.processAllCharCodes = processAllCharCodes;
		this.skipSpaceBar=skipSpaceBar;

		this.prevKeyCode;

		this.borderStyle = null;
		this.backgroundColor = null;
		this.color = null;
		// attach object's methods
		this.getNode = LK_TextNode.getNode;
		this.setUp = LK_TextNode.setUp;
		this.tearDown = LK_TextNode.tearDown;
		this.attachListener = LK_TextNode.attachListener;
		this.detachListener = LK_TextNode.detachListener;
	},

	getTranslationFunction : function() {
		return this.translationFunction;
	},

	getNode : function() {
		return this.txtnode;
	},

	setUp : function() {
		if (this.onKeyHandlerFunc!=null) {
			this.attachListener(this.eventToListen, this.onKeyHandlerFunc);
		} else {
			this.attachListener(this.eventToListen, this.translationFunction);
		}
		if (this.onCapsLockEventToListen!=null && this.onCapsLockHandlerFunc!=null) {
			this.attachListener(this.onCapsLockEventToListen, this.onCapsLockHandlerFunc);
		}
		for(var i=1;i<this.eventHandlerArray.length;i++) {
			this.attachListener(this.eventHandlerArray[i].eventType, this.eventHandlerArray[i].eventHandlerFunc);
		}
		// set some stupid style parameters.
		var style = "dashed 1px blue";
		if (this.isNodeEditor) {
			this.borderStyle = this.txtnode.style.border;
			this.txtnode.style.border = style;
		} else {
			this.borderStyle = this.txtnode.style.outline;
			this.txtnode.style.outline = style;
		}
		this.backgroundColor = document.defaultView.getComputedStyle(this.txtnode, "").getPropertyValue("background-color");
		this.txtnode.style.backgroundColor=getModifiedColor(this.backgroundColor, LK_TextNode.bgcOffsetRed,LK_TextNode.bgcOffsetGreen,LK_TextNode.bgcOffsetBlue);
		this.color = document.defaultView.getComputedStyle(this.txtnode, "").getPropertyValue("color");
		this.txtnode.style.color=getModifiedColor(this.color, LK_TextNode.fgcOffsetRed,LK_TextNode.fgcOffsetGreen,LK_TextNode.fgcOffsetBlue);
	},

	tearDown : function() {
		if (this.onKeyHandlerFunc!=null) {
			this.detachListener(this.eventToListen, this.onKeyHandlerFunc);
		} else {
			this.detachListener(this.eventToListen, this.translationFunction);
		}
		if (this.onCapsLockEventToListen!=null && this.onCapsLockHandlerFunc!=null) {
			this.detachListener(this.onCapsLockEventToListen, this.onCapsLockHandlerFunc);
		}
		for(var i=1;i<this.eventHandlerArray.length;i++) {
			this.detachListener(this.eventHandlerArray[i].eventType, this.eventHandlerArray[i].eventHandlerFunc);
		}
		// unset some stupid style parameters.
		this.txtnode.style.backgroundColor=this.backgroundColor;
		this.txtnode.style.color=this.color;
		if (this.isNodeEditor) {
			this.txtnode.style.border = this.borderStyle;
		} else {
			this.txtnode.style.outline = this.borderStyle;
		}
	},

	// attaches a listener to the txtnode, registers func to process, false - for bubbling
	attachListener : function(eventToListen, onKeyHandlerFunc) {
		if (eventToListen == undefined || onKeyHandlerFunc == undefined) {
			throw "Cannot attach listener.";
		}
		if (this.txtnode.contentDocument) {
			this.txtnode.contentDocument.documentElement.addEventListener(eventToListen, onKeyHandlerFunc, false);
		} else {
			this.txtnode.addEventListener(eventToListen, onKeyHandlerFunc, false);
		}
	},

	// detaches previously attached listener from the txtnode
	detachListener : function(eventToListen, onKeyHandlerFunc) {
		if (eventToListen == undefined || onKeyHandlerFunc == undefined) {
			throw "Cannot detach listener.";
		}
		if (this.txtnode.contentDocument) {
			this.txtnode.contentDocument.documentElement.removeEventListener(eventToListen, onKeyHandlerFunc, false);
		} else {
			this.txtnode.removeEventListener(eventToListen, onKeyHandlerFunc, false);
		}
	}
}